/*********************************************************************
 *
 * unixODBC Cursor Library
 *
 * Created by Nick Gorham
 * (nick@easysoft.com).
 *
 * copyright (c) 1999 Nick Gorham
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************
 *
 * $Id: SQLConnect.c,v 1.1 1999/09/19 22:22:50 ngorham Exp $
 *
 * $Log: SQLConnect.c,v $
 * Revision 1.1  1999/09/19 22:22:50  ngorham
 *
 *
 * Added first cursor library work, read only at the moment and only works
 * with selects with no where clause
 *
 *
 **********************************************************************/

#include "cursorlibrary.h"

static struct driver_func  cl_template_func[] =
{
    /* 00 */ { SQL_API_SQLALLOCCONNECT,      "SQLAllocConnect",
                            NULL },
    /* 01 */ { SQL_API_SQLALLOCENV,          "SQLAllocEnv",
                            NULL },
    /* 02 */ { SQL_API_SQLALLOCHANDLE,       "SQLAllocHandle", 
                            (SQLRETURN (*)())CLAllocHandle },
    /* 03 */ { SQL_API_SQLALLOCSTMT,         "SQLAllocStmt",
                            (SQLRETURN (*)())CLAllocStmt },
    /* 04 */ { SQL_API_SQLALLOCHANDLESTD,    "SQLAllocHandleStd",
                            (SQLRETURN (*)())CLAllocHandleStd },
    /* 05 */ { SQL_API_SQLBINDCOL,           "SQLBindCol",
                            (SQLRETURN (*)())CLBindCol },
    /* 06 */ { SQL_API_SQLBINDPARAM,         "SQLBindParam",
                            (SQLRETURN (*)())CLBindParam },
    /* 07 */ { SQL_API_SQLBINDPARAMETER,     "SQLBindParameter",
                            (SQLRETURN (*)())CLBindParameter },
    /* 08 */ { SQL_API_SQLBROWSECONNECT,     "SQLBrowseConnect",
                            NULL },
    /* 09 */ { SQL_API_SQLBULKOPERATIONS,    "SQLBulkOperations",
                            NULL },
    /* 10 */ { SQL_API_SQLCANCEL,            "SQLCancel",
                            (SQLRETURN (*)())CLCancel },
    /* 11 */ { SQL_API_SQLCLOSECURSOR,       "SQLCloseCursor",
                            (SQLRETURN (*)())CLCloseCursor },
    /* 12 */ { SQL_API_SQLCOLATTRIBUTE,      "SQLColAttribute",
                            (SQLRETURN (*)())CLColAttribute },
    /* 13 */ { SQL_API_SQLCOLATTRIBUTES,     "SQLColAttributes",
                            (SQLRETURN (*)())CLColAttributes },
    /* 14 */ { SQL_API_SQLCOLUMNPRIVILEGES,  "SQLColumnPrivileges",
                            (SQLRETURN (*)())CLColumnPrivileges },
    /* 15 */ { SQL_API_SQLCOLUMNS,           "SQLColumns",
                            (SQLRETURN (*)())CLColumns },
    /* 16 */ { SQL_API_SQLCONNECT,           "SQLConnect",
                            NULL },
    /* 17 */ { SQL_API_SQLCOPYDESC,          "SQLCopyDesc",
                            (SQLRETURN (*)())CLCopyDesc },
    /* 18 */ { SQL_API_SQLDATASOURCES,       "SQLDataSources",
                            NULL },
    /* 19 */ { SQL_API_SQLDESCRIBECOL,       "SQLDescribeCol",
                            (SQLRETURN (*)())CLDescribeCol },
    /* 20 */ { SQL_API_SQLDESCRIBEPARAM,     "SQLDescribeParam",
                            (SQLRETURN (*)())CLDescribeParam },
    /* 21 */ { SQL_API_SQLDISCONNECT,        "SQLDisconnect",
                            (SQLRETURN (*)())CLDisconnect },
    /* 22 */ { SQL_API_SQLDRIVERCONNECT,     "SQLDriverConnect",
                            NULL },
    /* 23 */ { SQL_API_SQLDRIVERS,           "SQLDrivers",
                            NULL },
    /* 24 */ { SQL_API_SQLENDTRAN,           "SQLEndTran",
                            (SQLRETURN (*)())CLEndTran },
    /* 25 */ { SQL_API_SQLERROR,             "SQLError",
                            (SQLRETURN (*)())CLError },
    /* 26 */ { SQL_API_SQLEXECDIRECT,        "SQLExecDirect",
                            (SQLRETURN (*)())CLExecDirect },
    /* 27 */ { SQL_API_SQLEXECUTE,           "SQLExecute",
                            (SQLRETURN (*)())CLExecute },
    /* 28 */ { SQL_API_SQLEXTENDEDFETCH,     "SQLExtendedFetch",
                            (SQLRETURN (*)())CLExtendedFetch },
    /* 29 */ { SQL_API_SQLFETCH,             "SQLFetch",
                            (SQLRETURN (*)())CLFetch },
    /* 30 */ { SQL_API_SQLFETCHSCROLL,       "SQLFetchScroll",
                            (SQLRETURN (*)())CLFetchScroll },
    /* 31 */ { SQL_API_SQLFOREIGNKEYS,       "SQLForeignKeys",
                            (SQLRETURN (*)())CLForeignKeys },
    /* 32 */ { SQL_API_SQLFREEENV,           "SQLFreeEnv",
                            NULL },
    /* 33 */ { SQL_API_SQLFREEHANDLE,        "SQLFreeHandle",
                            (SQLRETURN (*)())CLFreeHandle },
    /* 34 */ { SQL_API_SQLFREESTMT,          "SQLFreeStmt",
                            (SQLRETURN (*)())CLFreeStmt },
    /* 35 */ { SQL_API_SQLFREECONNECT,       "SQLFreeConnect",
                            NULL },
    /* 36 */ { SQL_API_SQLGETCONNECTATTR,    "SQLGetConnectAttr",
                            (SQLRETURN (*)())CLGetConnectAttr },
    /* 37 */ { SQL_API_SQLGETCONNECTOPTION,  "SQLGetConnectOption",
                            (SQLRETURN (*)())CLGetConnectOption },
    /* 38 */ { SQL_API_SQLGETCURSORNAME,     "SQLGetCursorName",
                            (SQLRETURN (*)())CLGetCursorName },
    /* 39 */ { SQL_API_SQLGETDATA,           "SQLGetData",
                            (SQLRETURN (*)())CLGetData },
    /* 40 */ { SQL_API_SQLGETDESCFIELD,      "SQLGetDescField",
                            (SQLRETURN (*)())CLGetDescField },
    /* 41 */ { SQL_API_SQLGETDESCREC,        "SQLGetDescRec",
                            (SQLRETURN (*)())CLGetDescRec },
    /* 42 */ { SQL_API_SQLGETDIAGFIELD,      "SQLGetDiagField",
                            (SQLRETURN (*)())CLGetDiagField },
    /* 43 */ { SQL_API_SQLGETENVATTR,        "SQLGetEnvAttr",
                            NULL },
    /* 44 */ { SQL_API_SQLGETFUNCTIONS,      "SQLGetFunctions",
                            NULL },
    /* 45 */ { SQL_API_SQLGETINFO,           "SQLGetInfo",
                            (SQLRETURN (*)())CLGetInfo },
    /* 46 */ { SQL_API_SQLGETSTMTATTR,       "SQLGetStmtAttr",
                            (SQLRETURN (*)())CLGetStmtAttr },
    /* 47 */ { SQL_API_SQLGETSTMTOPTION,     "SQLGetStmtOption",
                            (SQLRETURN (*)())CLGetStmtOption },
    /* 48 */ { SQL_API_SQLGETTYPEINFO,       "SQLGetTypeInfo",
                            (SQLRETURN (*)())CLGetTypeInfo },
    /* 49 */ { SQL_API_SQLMORERESULTS,       "SQLMoreResults",
                            (SQLRETURN (*)())CLMoreResults },
    /* 50 */ { SQL_API_SQLNATIVESQL,         "SQLNativeSql",
                            (SQLRETURN (*)())CLNativeSql },
    /* 51 */ { SQL_API_SQLNUMPARAMS,         "SQLNumParams",
                            (SQLRETURN (*)())CLNumParams },
    /* 52 */ { SQL_API_SQLNUMRESULTCOLS,     "SQLNumResultCols",
                            (SQLRETURN (*)())CLNumResultCols },
    /* 53 */ { SQL_API_SQLPARAMDATA,         "SQLParamData",
                            (SQLRETURN (*)())CLParamData },
    /* 54 */ { SQL_API_SQLPARAMOPTIONS,      "SQLParamOptions",
                            (SQLRETURN (*)())CLParamOptions },
    /* 55 */ { SQL_API_SQLPREPARE,           "SQLPrepare",
                            (SQLRETURN (*)())CLPrepare },
    /* 56 */ { SQL_API_SQLPRIMARYKEYS,       "SQLPrimaryKeys",
                            (SQLRETURN (*)())CLPrimaryKeys },
    /* 57 */ { SQL_API_SQLPROCEDURECOLUMNS,  "SQLProcedureColumns",
                            (SQLRETURN (*)())CLProcedureColumns },
    /* 58 */ { SQL_API_SQLPROCEDURES,        "SQLProcedures",
                            (SQLRETURN (*)())CLProcedures },
    /* 59 */ { SQL_API_SQLPUTDATA,           "SQLPutData",
                            (SQLRETURN (*)())CLPutData },
    /* 60 */ { SQL_API_SQLROWCOUNT,          "SQLRowCount",
                            (SQLRETURN (*)())CLRowCount },
    /* 61 */ { SQL_API_SQLSETCONNECTATTR,    "SQLSetConnectAttr",
                            (SQLRETURN (*)())CLSetConnectAttr },
    /* 62 */ { SQL_API_SQLSETCONNECTOPTION,  "SQLSetConnectOption",
                            (SQLRETURN (*)())CLSetConnectOption },
    /* 63 */ { SQL_API_SQLSETCURSORNAME,     "SQLSetCursorName",
                            (SQLRETURN (*)())CLSetCursorName },
    /* 64 */ { SQL_API_SQLSETDESCFIELD,      "SQLSetDescField",
                            (SQLRETURN (*)())CLSetDescField },
    /* 65 */ { SQL_API_SQLSETDESCREC,        "SQLSetDescRec", 
                            (SQLRETURN (*)())CLSetDescRec },
    /* 66 */ { SQL_API_SQLSETENVATTR,        "SQLSetEnvAttr",
                            NULL },
    /* 67 */ { SQL_API_SQLSETPARAM,          "SQLSetParam",
                            (SQLRETURN (*)())CLSetParam },
    /* 68 */ { SQL_API_SQLSETPOS,            "SQLSetPos",
                            (SQLRETURN (*)())CLSetPos },
    /* 69 */ { SQL_API_SQLSETSCROLLOPTIONS,  "SQLSetScrollOptions",
                            (SQLRETURN (*)())CLSetScrollOptions },
    /* 70 */ { SQL_API_SQLSETSTMTATTR,       "SQLSetStmtAttr",
                            (SQLRETURN (*)())CLSetStmtAttr },
    /* 71 */ { SQL_API_SQLSETSTMTOPTION,     "SQLSetStmtOption",
                            (SQLRETURN (*)())CLSetStmtOption },
    /* 72 */ { SQL_API_SQLSPECIALCOLUMNS,    "SQLSpecialColumns",
                            (SQLRETURN (*)())CLSpecialColumns },
    /* 73 */ { SQL_API_SQLSTATISTICS,        "SQLStatistics",
                            (SQLRETURN (*)())CLStatistics },
    /* 74 */ { SQL_API_SQLTABLEPRIVILEGES,   "SQLTablePrivileges",
                            (SQLRETURN (*)())CLTablePrivileges },
    /* 75 */ { SQL_API_SQLTABLES,            "SQLTables",
                            (SQLRETURN (*)())CLTables },
    /* 76 */ { SQL_API_SQLTRANSACT,          "SQLTransact",
                            (SQLRETURN (*)())CLTransact },
    /* 77 */ { SQL_API_SQLGETDIAGREC,        "SQLGetDiagRec",
                            (SQLRETURN (*)())CLGetDiagRec },
};

/*
 * connect is done by the driver manager, the is called to put the
 * cursor lib in the call chain
 */

SQLRETURN CLConnect( DMHDBC connection )
{
    int i;
    CLHDBC cl_connection;
    SQLCHAR s0[ 20 ];
    SQLRETURN ret;

    /*
     * Allocated a cursor connection structure 
     */

    cl_connection = malloc( sizeof( CLHDBC ));

    if ( !cl_connection )
    {
        logPushMsg( connection -> log_handle,
                __get_pid( s0 ),
                "CL " __FILE__,
                __LINE__,
                LOG_INFO,
                LOG_INFO,
                "Error: IM001" );

        __post_internal_error( &connection -> error,
                ERROR_HY001, NULL,
                connection -> environment -> requested_version );
        return SQL_ERROR;
    }

    memset( cl_connection, 0, sizeof( CLHDBC ));

    cl_connection -> functions = connection -> functions;
    cl_connection -> dm_connection = connection;

    /*
     * allocated a copy of the functions
     */

    if ( !( cl_connection -> functions = malloc( sizeof( cl_template_func ))))
    {
        logPushMsg( connection -> log_handle,
                __get_pid( s0 ),
                "CL " __FILE__,
                __LINE__,
                LOG_INFO,
                LOG_INFO,
                "Error: IM001" );

        __post_internal_error( &connection -> error,
                ERROR_HY001, NULL,
                connection -> environment -> requested_version );
        return SQL_ERROR;
    }

    /*
     * replace the function pointers with the ones in the
     * cursor lib
     */

    for ( i = 0; 
            i < sizeof( cl_template_func ) / sizeof( cl_template_func[ 0 ] ); 
            i ++ )
    {
        cl_connection -> functions[ i ] =
            connection -> functions[ i ];

        /*
         * if set replace the driver function with the function in 
         * the template
         */

        if ( cl_template_func[ i ].func && 
                connection -> functions[ i ].func )
        {
            connection -> functions[ i ] =
                cl_template_func[ i ];

            /*
             * copy the can_supply from the drivers list
             */

            connection -> functions[ i ].can_supply =
                cl_connection -> functions[ i ].can_supply;
        }
    }

    /*
     * add some functions the cursor lib will supply
     */

    connection -> functions[ DM_SQLSETPOS ].can_supply = 1;
    connection -> functions[ DM_SQLSETPOS ].func = 
        cl_template_func[ DM_SQLSETPOS ].func;

    connection -> functions[ DM_SQLSETSCROLLOPTIONS ].can_supply = 1;
    connection -> functions[ DM_SQLSETSCROLLOPTIONS ].func = 
        cl_template_func[ DM_SQLSETSCROLLOPTIONS ].func;

    connection -> functions[ DM_SQLFETCHSCROLL ].can_supply = 1;
    connection -> functions[ DM_SQLFETCHSCROLL ].func = 
        cl_template_func[ DM_SQLFETCHSCROLL ].func;

    connection -> functions[ DM_SQLEXTENDEDFETCH ].can_supply = 1;
    connection -> functions[ DM_SQLEXTENDEDFETCH ].func = 
        cl_template_func[ DM_SQLEXTENDEDFETCH ].func;

    /*
     * blank off what we don't do
     */

    connection -> functions[ DM_SQLBULKOPERATIONS ].can_supply = 0;
    connection -> functions[ DM_SQLBULKOPERATIONS ].func = NULL;

    /*
     * intercept the driver dbc
     */

    cl_connection -> driver_dbc = connection -> driver_dbc;
    connection -> driver_dbc = ( SQLHANDLE ) cl_connection;

    /*
     * check the number of alowed active statements
     */

    if ( CHECK_SQLGETINFO( cl_connection ))
    {
        ret = SQLGETINFO( cl_connection,
                cl_connection -> driver_dbc,
                SQL_MAX_CONCURRENT_ACTIVITIES,
                &cl_connection -> active_statement_allowed,
                0,
                NULL );
        /*
         * assume the worst
         */

        if ( !SQL_SUCCEEDED( ret ))
        {
            cl_connection -> active_statement_allowed = 1;
        }
    }
    else
    {
        cl_connection -> active_statement_allowed = 1;
    }

    return SQL_SUCCESS;
}

SQLRETURN CLDisconnect( SQLHDBC connection_handle )
{
    SQLRETURN ret;
    int i;
    CLHDBC cl_connection = (CLHDBC)connection_handle;
    DMHDBC connection = cl_connection -> dm_connection;

    /* 
     * disconnect from the driver
     */

    ret = SQLDISCONNECT( cl_connection, 
            cl_connection -> driver_dbc );

    if ( SQL_SUCCEEDED( ret ))
    {
        /*
         * replace the function pointers with the ones from the
         * cursor lib
         */

        for ( i = 0; 
            i < sizeof( cl_template_func ) / sizeof( cl_template_func[ 0 ] ); 
            i ++ )
        {
            connection -> functions[ i ] =
                cl_connection -> functions[ i ];
        }

        /*
         * replace the driver dbc
         */

        connection -> driver_dbc = 
            cl_connection -> driver_dbc;
    }

    return ret;
}

